/** Copyright (c) 2018 Uber Technologies, Inc.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 *
 * @flow
 */

/* eslint-env node */
import * as React from 'react';
import Stream from 'stream';
// $FlowFixMe - renderToPipeableStream
import {renderToString, renderToPipeableStream} from 'react-dom/server';
import type {Logger} from 'fusion-tokens';

export default (
  el: React.Element<*>,
  ctx: any,
  logger?: Logger,
  renderFullSSR?: (any) => boolean
): Promise<string> => {
  const renderFullSSRFn = renderFullSSR || ((ctx) => true);
  try {
    if (renderFullSSRFn(ctx)) {
      return Promise.resolve(`<div id='root'>${renderToString(el)}</div>`);
    } else {
      return new Promise((resolve, reject) => {
        ctx.res.socket.on('error', (error) => {
          console.error('Fatal server-side streaming error', error);
        });

        const shellTemplates = ctx.shellTemplate(ctx);

        // Create a stream
        const stream = new Stream.PassThrough();
        // Assign to koa
        ctx.body = stream;
        // Let React pipe to that same stream
        const {pipe} = renderToPipeableStream(<div id="root">{el}</div>, {
          nonce: ctx.nonce,
          bootstrapScripts: shellTemplates.scripts,
          onShellReady() {
            ctx.response.set('Content-Type', 'txt/html');
            ctx.type = 'html';
            stream.write(shellTemplates.start);
            pipe(stream);
            stream.write(shellTemplates.end);
          },
          onAllReady() {
            resolve('OK');
          },
          onError(e) {
            if (__DEV__) {
              console.error('Server-side stream failed', e);
            }
          },
        });
      });
    }
  } catch (e) {
    if (__DEV__) {
      console.error(
        'Server-side render failed. Falling back to client-side render',
        e
      );
    }
    logger && logger.error('SSR Failed with Error', e);
    return Promise.resolve('<div id="root" data-fusion-render="client"></div>');
  }
};
